@@ -1,13 +1,19 @@
 #include "ruby.h"
 
+#ifdef HAVE_RUBY_THREAD_H
+  #include "ruby/thread.h"
+#endif
+
 #ifdef MAKE_TRAP
 #include "rubysig.h"
 #endif
 
-
 #include <pcap.h>
+// Win32-Extensions only exist in winpcap not npcap
 #if defined(WIN32)
-#include <Win32-Extensions.h>
+  #ifdef HAVE_WIN32_EXTENSIONS_H
+    #include <Win32-Extensions.h>
+  #endif
 #endif
 
 #if !defined(WIN32)
@@ -18,7 +24,7 @@
 
 static VALUE mPCAP;
 static VALUE rb_cPcap, rb_cPkt;
-static VALUE ePCAPRUBError, eDumperError, eBindingError, eBPFilterError;
+static VALUE ePCAPRUBError, eDumperError, eBindingError, eBPFilterError, eLinkTypeError;
 
 #if defined(WIN32)
 static VALUE rbpcap_thread_wait_handle(HANDLE fno);
@@ -29,6 +35,7 @@ static VALUE rbpcap_thread_wait_handle(HANDLE fno);
 
 #define OFFLINE 1
 #define LIVE 2
+#define DEAD 3
 
 #if !defined(PCAP_NETMASK_UNKNOWN)
 /*
@@ -57,6 +64,15 @@ typedef struct rbpacket {
   u_char* pkt;
 } rbpacket_t;
 
+/*
+* Return the pcap version
+*/
+static VALUE
+rbpcap_s_lib_version(VALUE self)
+{
+  return rb_str_new2(pcap_lib_version());
+}
+
 /*
 * Return the name of a network device on the system.
 *
@@ -68,7 +84,6 @@ rbpcap_s_lookupdev(VALUE self)
   char *dev = NULL;
   char eb[PCAP_ERRBUF_SIZE];
   VALUE ret_dev;  /* device string to return */
-#if defined(WIN32)  /* pcap_lookupdev is broken on windows */
   pcap_if_t *alldevs;
   pcap_if_t *d;
 
@@ -91,13 +106,6 @@ rbpcap_s_lookupdev(VALUE self)
   ret_dev = rb_str_new2(dev);
   /* We don't need any more the device list. Free it */
   pcap_freealldevs(alldevs);
-#else
-  dev = pcap_lookupdev(eb);
-  if (dev == NULL) {
-	rb_raise(eBindingError, "%s", eb);
- }
-  ret_dev = rb_str_new2(dev);
-#endif
   return ret_dev;
 }
 
@@ -260,6 +268,73 @@ rbpcap_settimeout(VALUE self, VALUE timeout)
   }
 }
 
+/*
+* call-seq:
+*   listdatalinks()
+*
+* Get supported datalinks for this interface
+*
+* Returns an array of supported datalinks.
+*/
+static VALUE
+rbpcap_listdatalinks(VALUE self)
+{
+  rbpcap_t *rbp;
+  int *links, numlinks;
+  int i;
+  Data_Get_Struct(self, rbpcap_t, rbp);
+
+  numlinks = pcap_list_datalinks(rbp->pd, &links);
+  if (numlinks > 0) {
+    VALUE hash = rb_hash_new();
+    for (i = 0; i < numlinks; i++) {
+      const char *name = pcap_datalink_val_to_name(links[i]);
+      if (name != NULL) {
+        VALUE namestring = rb_str_new2(name);
+        rb_hash_aset(hash, INT2NUM(links[i]), namestring);
+      }
+    }
+    pcap_free_datalinks(links);
+    return hash;
+  } else {
+    rb_raise(eLinkTypeError, "unable to get datalinks (%d): %s", numlinks, pcap_geterr(rbp->pd));
+  }
+}
+
+
+/*
+* call-seq:
+*   setdatalink(1234)
+*
+* Set datalink type for the capture.
+*
+* Returns the object itself.
+*/
+static VALUE
+rbpcap_setdatalink(VALUE self, VALUE datalink)
+{
+  rbpcap_t *rbp;
+  int linktype, errcode;
+  Data_Get_Struct(self, rbpcap_t, rbp);
+
+  if(TYPE(datalink) == T_FIXNUM ) {
+    linktype = NUM2INT(datalink);
+  } else if (TYPE(datalink) == T_STRING) {
+    linktype = pcap_datalink_name_to_val(RSTRING_PTR(datalink));
+    if (linktype < 0) {
+      rb_raise(eLinkTypeError, "invalid datalink name: %s", RSTRING_PTR(datalink));
+    }
+  } else {
+    rb_raise(rb_eArgError, "datalink type must be a string or fixnum");
+  }
+
+  errcode = pcap_set_datalink(rbp->pd, linktype);
+  if (errcode == 0) {
+    return self;
+  } else {
+    rb_raise(eLinkTypeError, "unable to set datalink (%d): %s", errcode, pcap_geterr(rbp->pd));
+  }
+}
 
 /*
 * call-seq:
@@ -335,7 +410,7 @@ rbpcap_setfilter(VALUE self, VALUE filter)
 {
   char eb[PCAP_ERRBUF_SIZE];
   rbpcap_t *rbp;
-  u_int32_t mask = 0, netid = 0;
+  bpf_u_int32 mask = 0, netid = 0;
   struct bpf_program bpf;
 
   Data_Get_Struct(self, rbpcap_t, rbp);
@@ -343,7 +418,11 @@ rbpcap_setfilter(VALUE self, VALUE filter)
   if(TYPE(filter) != T_STRING)
   	rb_raise(eBPFilterError, "filter must be a string");
 
-	if(! rbpcap_ready(rbp)) return self;
+  if(! rbpcap_ready(rbp)) return self;
+
+  if(rbp->type == DEAD) {
+  	rb_raise(eBPFilterError, "unable to set bpf filter on OPEN_DEAD");
+  }
 
   if(rbp->type == LIVE)
   	if(pcap_lookupnet(rbp->iface, &netid, &mask, eb) < 0) {
@@ -352,12 +431,39 @@ rbpcap_setfilter(VALUE self, VALUE filter)
   		rb_warn("unable to get IP: %s", eb);
   	}
 
-  if(pcap_compile(rbp->pd, &bpf, RSTRING_PTR(filter), 0, mask) < 0)
-  	rb_raise(eBPFilterError, "invalid bpf filter");
+  if(pcap_compile(rbp->pd, &bpf, RSTRING_PTR(filter), 0, mask) < 0) {
+  	rb_raise(eBPFilterError, "invalid bpf filter: %s", pcap_geterr(rbp->pd));
+  }
 
-  if(pcap_setfilter(rbp->pd, &bpf) < 0)
-  	rb_raise(eBPFilterError, "unable to set bpf filter");
+  if(pcap_setfilter(rbp->pd, &bpf) < 0) {
+  	pcap_freecode(&bpf);
+  	rb_raise(eBPFilterError, "unable to set bpf filter: %s", pcap_geterr(rbp->pd));
+  }
 
+  pcap_freecode(&bpf);
+  return self;
+}
+
+/*
+* call-seq:
+*   compile(filter)
+*
+* Raises an exception if "filter" has a syntax error
+*
+* Returns self if the filter is valid
+*/
+static VALUE
+rbpcap_compile(VALUE self, VALUE filter) {
+  struct bpf_program bpf;
+  bpf_u_int32 mask = 0;
+  rbpcap_t *rbp;
+
+  Data_Get_Struct(self, rbpcap_t, rbp);
+  if(pcap_compile(rbp->pd, &bpf, RSTRING_PTR(filter), 0, mask) < 0) {
+    rb_raise(eBPFilterError, "invalid bpf filter: %s", pcap_geterr(rbp->pd));
+  }
+
+  pcap_freecode(&bpf);
   return self;
 }
 
@@ -565,7 +671,6 @@ rbpcap_open_dead(VALUE self, VALUE linktype, VALUE snaplen)
 {
   rbpcap_t *rbp;
 
-
   if(TYPE(linktype) != T_FIXNUM)
       rb_raise(rb_eArgError, "linktype must be a fixnum");
   if(TYPE(snaplen) != T_FIXNUM)
@@ -574,7 +679,7 @@ rbpcap_open_dead(VALUE self, VALUE linktype, VALUE snaplen)
   Data_Get_Struct(self, rbpcap_t, rbp);
 
   memset(rbp->iface, 0, sizeof(rbp->iface));
-  rbp->type = OFFLINE;
+  rbp->type = DEAD;
 
   rbp->pd = pcap_open_dead(
       NUM2INT(linktype),
@@ -793,6 +898,10 @@ rbpcap_next_data(VALUE self)
 	if(rbp->type == OFFLINE && ret <= 0)
 	  return Qnil;
 
+  if(rbp->type == DEAD && ret <= 0)
+    return Qnil;
+
+
 	if(ret > 0 && job.hdr.caplen > 0)
     return rb_str_new((char *) job.pkt, job.hdr.caplen);
 
@@ -838,6 +947,9 @@ rbpcap_next_packet(VALUE self)
 	if(rbp->type == OFFLINE && ret <= 0)
 	  return Qnil;
 
+  if(rbp->type == DEAD && ret <= 0)
+    return Qnil;
+
 	if(ret > 0 && job.hdr.caplen > 0)
     {
       rbpacket = ALLOC(rbpacket_t);
@@ -872,7 +984,7 @@ rbpcap_each_data(VALUE self)
   if(! rbpcap_ready(rbp)) return self;
 
 #if defined(WIN32)
-  fno = (int)pcap_getevent(rbp->pd);
+  fno = (HANDLE)pcap_getevent(rbp->pd);
 #else
   fno = pcap_get_selectable_fd(rbp->pd);
 #endif
@@ -880,6 +992,7 @@ rbpcap_each_data(VALUE self)
   for(;;) {
   	VALUE packet = rbpcap_next_data(self);
   	if(packet == Qnil && rbp->type == OFFLINE) break;
+    if(packet == Qnil && rbp->type == DEAD) break;
 #if defined(WIN32)
 	  packet == Qnil ? rbpcap_thread_wait_handle(fno) : rb_yield(packet);
 #else
@@ -913,7 +1026,7 @@ rbpcap_each_packet(VALUE self)
   if(! rbpcap_ready(rbp)) return self;
 
 #if defined(WIN32)
-  fno = (int)pcap_getevent(rbp->pd);
+  fno = (HANDLE)pcap_getevent(rbp->pd);
 #else
   fno = pcap_get_selectable_fd(rbp->pd);
 #endif
@@ -921,6 +1034,7 @@ rbpcap_each_packet(VALUE self)
   for(;;) {
   	VALUE packet = rbpcap_next_packet(self);
   	if(packet == Qnil && rbp->type == OFFLINE) break;
+    if(packet == Qnil && rbp->type == DEAD) break;
 #if defined(WIN32)
 	  packet == Qnil ? rbpcap_thread_wait_handle(fno) : rb_yield(packet);
 #else
@@ -1130,12 +1244,12 @@ rbpacket_data(VALUE self)
 }
 
 #if defined(WIN32)
-static VALUE
+static void *
 rbpcap_thread_wait_handle_blocking(void *data)
 {
   VALUE result;
   result = (VALUE)WaitForSingleObject(data, 100);
-  return result;
+  return (void *) result;
 }
 
 /*
@@ -1150,6 +1264,10 @@ rbpcap_thread_wait_handle(HANDLE fno)
 #if MAKE_TRAP
   // Ruby 1.8 doesn't support rb_thread_blocking_region
   result = rb_thread_polling();
+#elif defined(HAVE_RB_THREAD_CALL_WITHOUT_GVL)
+  result = (VALUE)rb_thread_call_without_gvl(
+      rbpcap_thread_wait_handle_blocking,
+	  fno, RUBY_UBF_IO, 0);
 #else
   result = (VALUE)rb_thread_blocking_region(
       rbpcap_thread_wait_handle_blocking,
@@ -1177,9 +1295,11 @@ Init_pcaprub_c()
   eBindingError = rb_path2class("PCAPRUB::BindingError");
   eBPFilterError = rb_path2class("PCAPRUB::BPFError");
   eDumperError = rb_path2class("PCAPRUB::DumperError");
+  eLinkTypeError = rb_path2class("PCAPRUB::LinkTypeError");
 
   rb_define_module_function(rb_cPcap, "lookupdev", rbpcap_s_lookupdev, 0);
   rb_define_module_function(rb_cPcap, "lookupnet", rbpcap_s_lookupnet, 1);
+  rb_define_module_function(rb_cPcap, "lib_version", rbpcap_s_lib_version, 0);
 
   rb_define_const(rb_cPcap, "DLT_NULL",   INT2NUM(DLT_NULL));
   rb_define_const(rb_cPcap, "DLT_EN10MB", INT2NUM(DLT_EN10MB));
@@ -1198,10 +1318,41 @@ Init_pcaprub_c()
   rb_define_const(rb_cPcap, "DLT_PPP_BSDOS", INT2NUM(DLT_PPP_BSDOS));
   rb_define_const(rb_cPcap, "DLT_IEEE802_11", INT2NUM(DLT_IEEE802_11));
   rb_define_const(rb_cPcap, "DLT_IEEE802_11_RADIO", INT2NUM(DLT_IEEE802_11_RADIO));
+#ifdef DLT_IEEE802_11_RADIO_AVS
   rb_define_const(rb_cPcap, "DLT_IEEE802_11_RADIO_AVS", INT2NUM(DLT_IEEE802_11_RADIO_AVS));
+#endif /* DLT_IEEE802_11_RADIO_AVS */
+#ifdef DLT_LINUX_SLL
   rb_define_const(rb_cPcap, "DLT_LINUX_SLL", INT2NUM(DLT_LINUX_SLL));
+#endif /* DLT_LINUX_SLL */
+#ifdef DLT_PRISM_HEADER
   rb_define_const(rb_cPcap, "DLT_PRISM_HEADER", INT2NUM(DLT_PRISM_HEADER));
+#endif /* DLT_PRISM_HEADER */
+#ifdef DLT_AIRONET_HEADER
   rb_define_const(rb_cPcap, "DLT_AIRONET_HEADER", INT2NUM(DLT_AIRONET_HEADER));
+#endif /* DLT_AIRONET_HEADER */
+
+#ifdef DLT_BLUETOOTH_HCI_H4
+  rb_define_const(rb_cPcap, "DLT_BLUETOOTH_HCI_H4", INT2NUM(DLT_BLUETOOTH_HCI_H4));
+#endif
+#ifdef DLT_BLUETOOTH_HCI_H4_WITH_PHDR
+  rb_define_const(rb_cPcap, "DLT_BLUETOOTH_HCI_H4_WITH_PHDR", INT2NUM(DLT_BLUETOOTH_HCI_H4_WITH_PHDR));
+#endif
+#ifdef DLT_IPFILTER
+  rb_define_const(rb_cPcap, "DLT_IPFILTER", INT2NUM(DLT_IPFILTER));
+#endif
+#ifdef DLT_LOOP
+  rb_define_const(rb_cPcap, "DLT_LOOP", INT2NUM(DLT_LOOP));
+#endif
+#ifdef DLT_USB
+  rb_define_const(rb_cPcap, "DLT_USB", INT2NUM(DLT_USB));
+#endif
+#ifdef DLT_USB_LINUX
+  rb_define_const(rb_cPcap, "DLT_USB_LINUX", INT2NUM(DLT_USB_LINUX));
+#endif
+#ifdef DLT_DOCSIS
+  rb_define_const(rb_cPcap, "DLT_DOCSIS", INT2NUM(DLT_DOCSIS));
+#endif
+
   /* Pcap Error Codes
    * Error codes for the pcap API.
    * These will all be negative, so you can check for the success or
@@ -1259,10 +1410,13 @@ Init_pcaprub_c()
 
   rb_define_method(rb_cPcap, "next", rbpcap_next_data, 0);
   rb_define_method(rb_cPcap, "setfilter", rbpcap_setfilter, 1);
+  rb_define_method(rb_cPcap, "compile", rbpcap_compile, 1);
   rb_define_method(rb_cPcap, "setmonitor", rbpcap_setmonitor, 1);
   rb_define_method(rb_cPcap, "setsnaplen", rbpcap_setsnaplen, 1);
   rb_define_method(rb_cPcap, "settimeout", rbpcap_settimeout, 1);
   rb_define_method(rb_cPcap, "setpromisc", rbpcap_setpromisc, 1);
+  rb_define_method(rb_cPcap, "setdatalink", rbpcap_setdatalink, 1);
+  rb_define_method(rb_cPcap, "listdatalinks", rbpcap_listdatalinks, 0);
   rb_define_method(rb_cPcap, "activate", rbpcap_activate, 0);
   rb_define_method(rb_cPcap, "inject", rbpcap_inject, 1);
   rb_define_method(rb_cPcap, "datalink", rbpcap_datalink, 0);
